#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/win32/tadsimg.cpp,v 1.2 1999/05/17 02:52:25 MJRoberts Exp $";
#endif

/* Copyright (c) 1997 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  tadsimg.cpp - TADS image base class
Function
  
Notes
  
Modified
  11/08/97 MJRoberts  - Creation
*/

#include <windows.h>

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef TADSWIN_H
#include "tadswin.h"
#endif
#ifndef TADSIMG_H
#include "tadsimg.h"
#endif


CTadsImage::CTadsImage()
{
    /* no pixel buffer yet */
    pix_ = 0;

    /* no mask buffer yet */
    mask_ = 0;

    /* no palette yet */
    palhdl_ = 0;
    colors_ = 0;

    /* no dimensions yet */
    width_ = 0;
    height_ = 0;

    /* presume we'll use the current display's bits per pixel */
    bpp_ = 0;
}

CTadsImage::~CTadsImage()
{
    /* delete any existing image */
    delete_image();
}

/*
 *   delete the image 
 */
void CTadsImage::delete_image()
{
    /* if there's a palette, delete it */
    if (palhdl_ != 0)
        DeleteObject(palhdl_);

    /* if we have a pixel buffer, delete it */
    if (pix_ != 0)
        os_free_huge(pix_);

    /* if we have a mask buffer, delete it */
    if (mask_ != 0)
        os_free_huge(mask_);

    /* if we have a color list, delete it */
    if (colors_ != 0)
        delete [] colors_;

    /* clear members */
    palhdl_ = 0;
    pix_ = 0;
    width_ = 0;
    height_ = 0;
}

/*
 *   Create a system palette from an array of HTML_color_t's 
 */
void CTadsImage::create_palhdl(HTML_color_t *pal, size_t palcnt)
{
    LOGPALETTE *logpal;
    PALETTEENTRY *entry;
    size_t indx;
    
    /* if we have a palette already, delete it */
    if (palhdl_ != 0)
    {
        DeleteObject(palhdl_);
        palhdl_ = 0;
    }

    /* if there's no array, do nothing */
    if (pal == 0)
        return;

    /* allocate a LOGPALETTE structure with space for the colors */
    logpal = (LOGPALETTE *)th_malloc(sizeof(LOGPALETTE)
                                     + palcnt * sizeof(PALETTEENTRY));

    /* set up the structure header */
    logpal->palVersion = 0x0300;
    logpal->palNumEntries = (WORD)palcnt;

    /* store a copy of the color list */
    colors_ = new HTML_color_t[palcnt];
    color_cnt_ = palcnt;
    memcpy(colors_, pal, palcnt * sizeof(*colors_));

    /* set up the color entries */
    for (indx = 0, entry = &logpal->palPalEntry[0] ; indx < palcnt ;
         ++indx, ++entry, ++pal)
    {
        /* store the palette entry */
        entry->peRed = HTML_color_red(*pal);
        entry->peGreen = HTML_color_green(*pal);
        entry->peBlue = HTML_color_blue(*pal);
        entry->peFlags = 0;
    }

    /* create our palette */
    palhdl_ = CreatePalette(logpal);
    UnrealizeObject(palhdl_);

    /* done with the logical palette structure */
    th_free(logpal);
}


/*
 *   Draw the image 
 */
void CTadsImage::draw(CTadsWin *win, CHtmlRect *pos,
                      tadsimg_draw_mode_t mode)
{
    int use_pal;
    BITMAPINFO *bmi;
    HPALETTE oldpal;
    HPALETTE oldpal_dc;
    HPALETTE sys_pal = 0;
    HDC hdc;
    HTML_color_t *pal;
    size_t i;
    long ht, wid;
    int blt_mode;

    /* use the palette if we're in 256-color mode or lower */
    use_pal = (win->get_bits_per_pixel() <= 8);

    /* don't do anything if not drawing */
    if ((hdc = win->gethdc()) == 0)
        return;

    /* presume we'll draw in 'source copy' mode */
    blt_mode = SRCCOPY;

    /* if we have a mask, draw it first */
    if (mask_ != 0)
    {
        /* 
         *   AND the mask to set background pixels to black wherever we're
         *   drawing opaque image pixels 
         */
        draw_mask(hdc, win, pos, mode);

        /* set the drawing mode to OR for the image */
        blt_mode = SRCPAINT;
    }

    /* 
     *   allocate the BITMAPINFO structure, including space for the
     *   palette if we're using a palette 
     */
    bmi = (BITMAPINFO *)th_malloc(sizeof(BITMAPINFO)
                                  + (use_pal
                                     ? color_cnt_*sizeof(RGBQUAD) : 0));

    /* fill in the palette, if we have one */
    if (colors_ != 0)
    {
        for (i = 0, pal = colors_ ; i < color_cnt_ ; ++i, ++pal)
        {
            bmi->bmiColors[i].rgbBlue = HTML_color_blue(*pal);
            bmi->bmiColors[i].rgbGreen = HTML_color_green(*pal);
            bmi->bmiColors[i].rgbRed = HTML_color_red(*pal);
        }
    }

    /* fill in the bitmap info header */
    bmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmi->bmiHeader.biWidth = width_;
    bmi->bmiHeader.biHeight = height_;
    bmi->bmiHeader.biPlanes = 1;
    bmi->bmiHeader.biCompression = use_pal ? 0 : BI_RGB;
    bmi->bmiHeader.biSizeImage = 0;
    bmi->bmiHeader.biXPelsPerMeter = 0;
    bmi->bmiHeader.biYPelsPerMeter = 0;
    bmi->bmiHeader.biClrUsed = (use_pal && colors_ != 0) ? color_cnt_ : 0;
    bmi->bmiHeader.biClrImportant = (use_pal && colors_ != 0)
                                    ? color_cnt_ : 0;

    /* 
     *   set the bits per pixel to the setting stored in the image - if
     *   bpp_ is zero, though, assume that the image is using the same
     *   depth as the display 
     */
    bmi->bmiHeader.biBitCount = (WORD)(bpp_ != 0
                                       ? bpp_
                                       : (use_pal ? 8 : 24));

    /* 
     *   Select and realize our palette, if using one.  Note that we
     *   select the palette as a background palette, since we take care of
     *   actually creating the system palette at window activation time or
     *   when creating a new image; now we only need to line up our colors
     *   with the ones in the current palette.  
     */
    if (use_pal && palhdl_ != 0)
    {
        /* 
         *   select the palette into the display device context and
         *   realize the palette - note that we must select the palette
         *   into the actual display device so that it gets mapped into
         *   the system palette 
         */
        oldpal = SelectPalette(win->get_display_hdc(), palhdl_, TRUE);
        RealizePalette(win->get_display_hdc());

        /* 
         *   if the drawing device context is different than the display
         *   device context (which usually means we're drawing into a
         *   memory DC), select the palette into the drawing DC as well,
         *   so that we map color indices for the drawing DC 
         */
        if (hdc != win->get_display_hdc())
        {
            struct
            {
                LOGPALETTE logpal;
                PALETTEENTRY entry[255];
            } pal_info;
            int i;
            int reserved;

            /* get the reserved palette entry count */
            reserved = GetDeviceCaps(win->get_display_hdc(), NUMRESERVED);

            /* get the system palette colors */
            GetSystemPaletteEntries(win->get_display_hdc(), 0, 256,
                                    &pal_info.logpal.palPalEntry[0]);

            /* set the color count in the palette */
            pal_info.logpal.palNumEntries = 256;

            /* clear all of the palette entry flags */
            for (i = 0 ; i < 256 ; ++i)
                pal_info.logpal.palPalEntry[i].peFlags = 0;

            /* create a logical palette containing the system colors */
            pal_info.logpal.palVersion = 0x0300;
            sys_pal = CreatePalette(&pal_info.logpal);

            /* select the system color palette into the memory DC */
            oldpal_dc = SelectPalette(hdc, sys_pal, FALSE);
            RealizePalette(hdc);
        }
        else
        {
            /* we don't have a memory DC, so there's no separate palette */
            oldpal_dc = 0;
        }
    }
    else
    {
        /* we're not using a palette */
        oldpal = 0;
        oldpal_dc = 0;
    }

    /* figure out how we're going to draw the image into the given area */
    switch(mode)
    {
    case TADSIMG_DRAW_CLIP:
        /* 
         *   draw at exact size. aligning at upper left of destination
         *   area, and clipping if it's too big 
         */
        wid = width_ < (unsigned long)(pos->right - pos->left)
              ? width_ : pos->right - pos->left;
        ht = height_ < (unsigned long)(pos->bottom - pos->top)
             ? height_ : pos->bottom - pos->top;

        /* display the bitmap */
        StretchDIBits(hdc,
                      pos->left, pos->top, wid, ht,
                      0, 0, wid, ht,
                      pix_,
                      (LPBITMAPINFO)&bmi->bmiHeader,
                      DIB_RGB_COLORS, blt_mode);
        break;

    case TADSIMG_DRAW_STRETCH:
        /*
         *   stretch the image to fill the given area 
         */
        StretchDIBits(hdc, pos->left, pos->top,
                      pos->right - pos->left, pos->bottom - pos->top,
                      0, 0, width_, height_,
                      pix_,
                      (LPBITMAPINFO)&bmi->bmiHeader,
                      DIB_RGB_COLORS, blt_mode);
        break;

    case TADSIMG_DRAW_TILE:
        /*
         *   draw at the exact size as many times as necessary to tile the
         *   area 
         */
        {
            long curx, cury;

            for (curx = pos->left ; curx < pos->right ; curx += width_)
            {
                for (cury = pos->top ; cury < pos->bottom ; cury += height_)
                {
                    /* 
                     *   draw the image at the current tile position,
                     *   clipping if necessary 
                     */
                    wid = width_ < (unsigned long)(pos->right - curx)
                          ? width_ : pos->right - curx;
                    ht = height_ < (unsigned long)(pos->bottom - cury)
                         ? height_ : pos->bottom - cury;

                    /* display the bitmap */
                    StretchDIBits(hdc,
                                  curx, cury, wid, ht,
                                  0, 0, wid, ht,
                                  pix_,
                                  (LPBITMAPINFO)&bmi->bmiHeader,
                                  DIB_RGB_COLORS, blt_mode);
                }
            }
        }
        break;
    }

    /* restore the old palette */
    if (oldpal != 0)
    {
        /* restore the original display device palette */
        SelectPalette(win->get_display_hdc(), oldpal, TRUE);
        RealizePalette(win->get_display_hdc());
    }

    /* restore the old memory device palette as well */
    if (oldpal_dc != 0)
    {
        /* restore the old memory dc palette */
        SelectPalette(hdc, oldpal_dc, FALSE);

        /* delete the system palette we created */
        DeleteObject(sys_pal);
    }

    /* done with the bitmapinfo structure */
    th_free(bmi);
}

/*
 *   Draw the mask
 */
void CTadsImage::draw_mask(HDC hdc, CTadsWin *win, CHtmlRect *pos,
                           tadsimg_draw_mode_t mode)
{
    struct
    {
        BITMAPINFO bmi;
        RGBQUAD clr;
    }
    bmi;
    long ht, wid;

    /* fill in the bitmap info header */
    bmi.bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmi.bmi.bmiHeader.biWidth = width_;
    bmi.bmi.bmiHeader.biHeight = height_;
    bmi.bmi.bmiHeader.biPlanes = 1;
    bmi.bmi.bmiHeader.biCompression = 0;
    bmi.bmi.bmiHeader.biSizeImage = 0;
    bmi.bmi.bmiHeader.biXPelsPerMeter = 0;
    bmi.bmi.bmiHeader.biYPelsPerMeter = 0;
    bmi.bmi.bmiHeader.biClrUsed = 2;
    bmi.bmi.bmiHeader.biClrImportant = 2;
    bmi.bmi.bmiHeader.biBitCount = 1;

    /* fill in the two-color palette - just black and white */
    bmi.bmi.bmiColors[0].rgbBlue = 0x00;
    bmi.bmi.bmiColors[0].rgbGreen = 0x00;
    bmi.bmi.bmiColors[0].rgbRed = 0x00;

    bmi.bmi.bmiColors[1].rgbBlue = 0xff;
    bmi.bmi.bmiColors[1].rgbGreen = 0xff;
    bmi.bmi.bmiColors[1].rgbRed = 0xff;

    /* figure out how we're going to draw the image into the given area */
    switch(mode)
    {
    case TADSIMG_DRAW_CLIP:
        /* 
         *   draw at exact size. aligning at upper left of destination
         *   area, and clipping if it's too big 
         */
        wid = width_ < (unsigned long)(pos->right - pos->left)
              ? width_ : pos->right - pos->left;
        ht = height_ < (unsigned long)(pos->bottom - pos->top)
             ? height_ : pos->bottom - pos->top;

        /* display the bitmap */
        StretchDIBits(hdc,
                      pos->left, pos->top, wid, ht,
                      0, 0, wid, ht,
                      mask_,
                      (LPBITMAPINFO)&bmi.bmi.bmiHeader,
                      DIB_RGB_COLORS,
                      SRCAND);
        break;

    case TADSIMG_DRAW_STRETCH:
        /*
         *   stretch the image to fill the given area 
         */
        StretchDIBits(hdc, pos->left, pos->top,
                      pos->right - pos->left, pos->bottom - pos->top,
                      0, 0, width_, height_,
                      mask_,
                      (LPBITMAPINFO)&bmi.bmi.bmiHeader,
                      DIB_RGB_COLORS,
                      SRCAND);
        break;

    case TADSIMG_DRAW_TILE:
        /*
         *   draw at the exact size as many times as necessary to tile the
         *   area 
         */
        {
            long curx, cury;

            for (curx = pos->left ; curx < pos->right ; curx += width_)
            {
                for (cury = pos->top ; cury < pos->bottom ; cury += height_)
                {
                    /* 
                     *   draw the image at the current tile position,
                     *   clipping if necessary 
                     */
                    wid = width_ < (unsigned long)(pos->right - curx)
                          ? width_ : pos->right - curx;
                    ht = height_ < (unsigned long)(pos->bottom - cury)
                         ? height_ : pos->bottom - cury;

                    /* display the bitmap */
                    StretchDIBits(hdc,
                                  curx, cury, wid, ht,
                                  0, 0, wid, ht,
                                  mask_,
                                  (LPBITMAPINFO)&bmi.bmi.bmiHeader,
                                  DIB_RGB_COLORS,
                                  SRCAND);
                }
            }
        }
        break;
    }
}


/*
 *   Map the image's palette into the system palette 
 */
int CTadsImage::map_palette(CTadsWin *win, int foreground)
{
    int ret;

    /* if we have a palette, map it */
    if (palhdl_ != 0)
    {
        HPALETTE oldpal;
        HDC dc;

        /* get the device context for the window */
        dc = GetDC(win->get_handle());

        /* select and realize the palette */
        oldpal = SelectPalette(dc, palhdl_, !foreground);
        ret = RealizePalette(dc);

        /* restore the old palette in the background */
        SelectPalette(dc, oldpal, TRUE);
        RealizePalette(dc);

        /* done with the window's DC */
        ReleaseDC(win->get_handle(), dc);

        /* return the color change indication */
        return ret;
    }
    else
    {
        /* no palette -> no changes */
        return 0;
    }
}

